#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <fcntl.h>
#include "sbig.h"

static struct sbig_init  info;

static int write_fits(char *filename, int width, int height, 
	unsigned short *data);

static double getf(char *s, float default_value)
{
    char buffer[80];

    printf(s);
    fflush(stdout);
    if (fgets(buffer, 79, stdin) == NULL)
	exit(0);
    if (buffer[0] == '\n')
	return default_value;
    return atof(buffer);
}

static int geti(char *s, int default_value)
{
    char buffer[80];

    printf(s);
    fflush(stdout);
    if (fgets(buffer, 79, stdin) == NULL)
	exit(0);
    if (buffer[0] == '\n')
	return default_value;
    return atoi(buffer);
}


/*
 * This is the 'progress' callback
 * passed to sbig_readout()
 */
static int readout_callback(float percent)
{
    printf("\r%4.1f%%  ", percent);

    /* return 1 to continue, 0 to abort */
    return 1;
}

/*
 *  Start an exposure, wait for it, then download from the camera
 *  in bands.  This is somewhat complex, since it handles any size
 *  subimage of the CCD, and bands too.
 */
static void take_picture(void)
{
    int  ret, i, needed;
    char  filename[80], input[80];
    static unsigned short  *buffer;
    static int  buffer_size;
    struct sbig_expose  expose;
    struct sbig_status  status;
    struct sbig_readout readout;
    static int  id = 0;
    double  f;

    /* loop until we get good input */
    while (1) {
	expose.ccd = geti("ccd? [*0=imaging, 1=tracking] ", 0);
	if (expose.ccd == 0)
	    readout.binning = geti("Binning mode [0=1x1, *1=2x2, 2=3x3]: ", 1);
	else if (expose.ccd == 1)
	    readout.binning = geti("Binning mode [0=1x1, *1=2x2]: ", 1);
	else {
	    printf("Invalid ccd %d\n", expose.ccd);
	    continue;
	}
	if (readout.binning < 0 || readout.binning > (expose.ccd == 0? 2:1)) {
	    printf("Invalid binning mode %d\n", readout.binning);
	    continue;
	}
	break;
    }

    /* loop until we get good input */
    while (1) {
	printf("Enter x, y, w, h for FITS image [CR for default]: ");
	fflush(stdout);
	if (fgets(input, 79, stdin) == NULL)
	    exit(0);
	if (input[0] == '\n') {
	    readout.x = readout.y = 0;
	    readout.width =
	      info.camera_info[expose.ccd].readout_mode[readout.binning].width;
	    readout.height =
	      info.camera_info[expose.ccd].readout_mode[readout.binning].height;
	    printf("image size is %d x %d\n", readout.width, readout.height);
	} else
	    sscanf(input, "%d %d %d %d", &readout.x, &readout.y,
		   &readout.width, &readout.height);
	if (readout.x < 0 || readout.y < 0) {
	    printf("invalid starting coordinate %d,%d\n",
		    readout.x, readout.y);
	    continue;
	}
	if (readout.width <= 0 || readout.height <= 0) {
	    printf("invalid width or height %dx%d\n", readout.width, readout.height);
	    continue;
	}
	break;
    }

    f = getf("Enter exposure time (sec) [default 5.0]: ", 5.0);
    expose.exposure_time = (int)(100.0*f);
    expose.abg_state = 2;

    i = geti("Expose image (*0), or DarkFrame (1): ", 0);
    expose.shutter = (i != 1);

    sync();
    ret = sbig_expose(&expose);
    printf("sbig_expose returned %d\n", ret);
    if (ret < 0)
	printf("sbig error: %s\n", sbig_show_error(ret));

    printf("Sleeping for %g seconds\n", 0.01*expose.exposure_time);
    if (expose.exposure_time > 100) {
	for (i = 0; i < expose.exposure_time; i += 100) {
	    sbig_get_status(&status);
	    printf("imaging_ccd_status=%d, shutter_state=%d\n",
		status.imaging_ccd_status, status.shutter_state);
	    sleep(1);
	}
    } else
	usleep(expose.exposure_time*10000);

    /* wait for exposure to complete */
    printf("Checking exposure\n");
    for (i = 0; i < 1000; ++i) {
	ret = sbig_get_status(&status);
	if (ret < 0) {
	    printf("sbig error: %s\n", sbig_show_error(ret));
	    break;
	}
	printf("ccd_status=%d\n", status.imaging_ccd_status);
	if (readout.ccd == 0) {
	    if (status.imaging_ccd_status != 2)
		break;
	} else {
	    if (status.tracking_ccd_status != 2)
		break;
	}
	usleep(50000);
    }
    printf("took an extra %g seconds\n", 0.050*i);
    if (i == 1000)
	printf("EXPOSURE DIDN'T FINISH!\n");
    printf("Exposure Complete\n");

    readout.ccd = expose.ccd;
    readout.binning = readout.binning;
    needed = readout.height*readout.width*sizeof(short);
    if (buffer_size < needed) {
	if (buffer != NULL)
	    free(buffer);
	buffer = malloc(buffer_size = needed);
    }
    readout.data = buffer;
    readout.data_size_in_bytes = buffer_size;
    readout.callback = readout_callback;

    sync();
    if ((ret = sbig_readout(&readout)) < 0) {
	printf("sbig error: %s\n", sbig_show_error(ret));
	return;
    }
    sprintf(filename, "st7%d.fits", id++);
    printf("\nWriting %s\n", filename);
    write_fits(filename, readout.width, readout.height, buffer);
    printf(" -- Done\n");
}

static void write_fits_record(int fd, char *s)
{
    register int  i;
    char  b[80];

    for (i = 0; i < 80; ++i)
        b[i] = (*s)? *s++ : ' ';
    write(fd, b, 80);
}
#define W(x) write_fits_record(fd, x)

static int write_fits(char *filename, int w, int h, unsigned short *data)
{
    register int i;
    int  min, max;
    int fd;
    char b[128];
    unsigned short u;

    if ((fd = open(filename, O_CREAT|O_WRONLY, 0666)) < 0) {
	printf("Cannot open '%s' for writing", filename);
	return -1;
    }
	    /*           1         2         3 */
            /*  123456789012345678901234567890 */
    W(         "SIMPLE  =                    T  / conforms to FITS");
    W(         "BITPIX  =                   16  / 16-bit twos comp integer");
    W(         "NAXIS   =                    2  / two axes");
    sprintf(b, "NAXIS1  =                  %03d", w); W(b);
    sprintf(b, "NAXIS2  =                  %03d", h); W(b);
    W("END"); 
    for (i = 6; i < 36; ++i)
	W(" ");
    /* shouldn't we scale this to fit, first? */
    min = 10000000;
    max = -10000000;
    for (i = w*h; --i >= 0; )
    {
	u = data[i];
	if (u > max)
	    max = u;
	if (u < min)
	    min = u;
	data[i] = (u<<8)|(u>>8);
    }
    printf("Wrote %d bytes of data, min=%d, max=%d\n", w*h*2, min, max);
    write(fd, data, w*h*sizeof(unsigned short));
    close(fd);
    sync();
    return 0;
}

int main(int argc, char *argv[])
{
    char buffer[80], last_command[80];
    int  ret, i, port;
    int  is_small_camera;

    port = 1; /* 0x378 */
    for (i = 1; i < argc; ++i) {
	if (!strcmp(argv[i], "-port")) {
	    sscanf(argv[++i], "%x", &port);
	    printf("port override to 0x%x\n", port);
	}
    }

    for (i = 0; i < 4; ++i) {
	sync();
	/* set our verbosity to 5 (range is 0 to 5) */
	ret = sbig_init(port, 5, &info);
	printf("sbig_init() returned %d\n", ret);
	if (ret < 0)
	    printf("sbig error: %s\n", sbig_show_error(ret));
	else
	    break;
	usleep(100000);
    }
    if (ret < 0) {
	printf("cannot open camera, giving up\n");
	exit(1);
    }
    printf("linux_version: %g\n", info.linux_version);
    printf("firmware_version: %u\n", info.firmware_version);
    printf("Imaging ABG: %s\n", info.imaging_abg_type? "Present":"None");
    printf("Imaging bad columns: %d\n", info.nmbr_bad_columns);
    for (i = 0; i < info.nmbr_bad_columns; ++i)
	printf("  column %d\n", info.bad_columns[i]);
    printf("Serial Number: %s\n", info.serial_number);
    printf("Camera name: %s\n", info.camera_name);
    is_small_camera = (strstr(info.camera_name, "237") != NULL ||
       index(info.camera_name, '5') != NULL);
    for (i = 0; i < info.camera_info[0].nmbr_readout_modes; ++i) {
	struct readout_mode  *rmp;
	rmp = &info.camera_info[0].readout_mode[i];
	printf("ImagingCCD mode %d: w=%d h=%d ", rmp->mode, rmp->width,
		rmp->height);
	printf("gain=%f ", 0.01*rmp->gain);
	printf("pw=%g ph=%g\n", 0.001*rmp->pixel_width, 0.001*rmp->pixel_height);
    }
    for (i = 0; i < info.camera_info[1].nmbr_readout_modes; ++i) {
	struct readout_mode  *rmp;
	rmp = &info.camera_info[1].readout_mode[i];
	printf("TrackingCCD mode %d: w=%d h=%d ", rmp->mode, rmp->width,
		rmp->height);
	printf("gain=%f ", 0.01*rmp->gain);
	printf("pw=%g ph=%g\n", 0.001*rmp->pixel_width, 0.001*rmp->pixel_height);
    }
    printf("ST5_AD_size = %d\n", info.ST5_AD_size);
    printf("ST5_filter_type = %d\n", info.ST5_filter_type);

    strcpy(last_command, "?");
    while(1) {
	printf("> ");
	fflush(stdout);
	if (fgets(buffer, 79, stdin) == NULL)
	    exit(0);
	if (buffer[0] == '\0')
	    strcpy(buffer, last_command);
	if (buffer[0] == 'q')
	    break;
	if (buffer[0] == '?') {
	    printf("status\n");
	    printf("quit\n");
	    printf("filter\n");
	    printf("misc\n");
	    printf("pulse\n");
	    printf("relay\n");
	    printf("cool\n");
	    printf("expose\n");
	    continue;
	}
	strcpy(last_command, buffer);

	if (strcmp(buffer, "status\n") == 0) {
	    struct sbig_status  stat;
	    sync();
	    ret = sbig_get_status(&stat);
	    printf("sbig_status returned %d\n", ret);
	    if (ret < 0) {
		printf("sbig error: %s\n", sbig_show_error(ret));
		continue;
	    }
	    printf("imaging_ccd_status=%d\n", stat.imaging_ccd_status);
	    printf("tracking_ccd_status=%d\n", stat.tracking_ccd_status);
	    printf("fan_on=%d\n", stat.fan_on);
	    printf("shutter_state=%d\n", stat.shutter_state);
	    printf("led_state=%d\n", stat.led_state);
	    printf("shutter_edge=%d\n", stat.shutter_edge);
	    printf("plus_x_relay=%d\n", stat.plus_x_relay);
	    printf("minus_x_relay=%d\n", stat.minus_x_relay);
	    printf("plus_y_relay=%d\n", stat.plus_y_relay);
	    printf("minus_y_relay=%d\n", stat.minus_y_relay);
	    printf("pulse_active=%d\n", stat.pulse_active);
	    printf("temperature_regulation=%d\n", stat.temperature_regulation);
	    printf("temperature_setpoint=%d\n", stat.temperature_setpoint);
	    printf("cooling_power=%d\n", stat.cooling_power);
	    printf("air_temperature=%d.%01d\n", stat.air_temperature/10,
			stat.air_temperature%10);
	    printf("ccd_temperature=%d.%01d\n", stat.ccd_temperature/10,
			stat.air_temperature%10);
	    continue;
	}
	if (strcmp(buffer, "expose\n") == 0) {
	    take_picture();
	    continue;
	}
	if (strcmp(buffer, "misc\n") == 0) {
	    struct sbig_control  ctrl;
	    ctrl.fan_on = geti("fan? [0 OFF, *ON 1]: ", 1);
	    ctrl.shutter = geti("shutter [0 *NOCHANGE, 1 OPEN, "
		" 2 CLOSE, 3 REINIT]: ", 0);
	    ctrl.led = geti("LED [0 OFF, 1 *ON, 2 SLOW, 3 FAST]: ", 1);
	    sync();
	    ret = sbig_control(&ctrl);
	    printf("sbig_control() returned %d\n", ret);
	    if (ret < 0)
		printf("sbig error: %s\n", sbig_show_error(ret));
	    continue;
	}
	if (!strcmp(buffer, "relay\n")) {
	    struct sbig_relay  relay;
	    i = geti("X time? [+- 0.1 s] ", 0);
	    if (i < 0) {
		relay.x_plus_time = 0;
		relay.x_minus_time = -i;
	    } else {
		relay.x_plus_time = i;
		relay.x_minus_time = 0;
	    }
	    i = geti("Y time? [+- 0.1 s] ", 0);
	    if (i < 0) {
		relay.y_plus_time = 0;
		relay.y_minus_time = -i;
	    } else {
		relay.y_plus_time = i;
		relay.y_minus_time = 0;
	    }
	    sync();
	    ret = sbig_activate_relay(&relay);
	    printf("sbig_activate_relay() returned %d\n", ret);
	    if (ret < 0)
		printf("sbig error: %s\n", sbig_show_error(ret));
	    continue;
	}
	if (strcmp(buffer, "filter\n") == 0) {
	    struct sbig_pulse  pulse;
	    if (is_small_camera) {
		i = geti("filter 1(open), 2(closed), 3(R), 4(G), 5(B) "
		  " [1 to 5]: ", 0);
	    } else {
		i = geti("CFW8 position [1 to 5]: ", 0);
	    }
	    if (i < 1 || i > 5) {
		printf("filter position is 1, 2, 3, 4 or 5\n");
		continue;
	    }
	    if (is_small_camera) {
		pulse.nmbr_pulses = 1;
		pulse.pulse_width = 0;
		pulse.pulse_interval = i;
	    } else {
		pulse.pulse_interval = 18270;
		pulse.nmbr_pulses = 60;
		pulse.pulse_width = 500 + 300*(i-1);
	    }
	    sync();
	    ret = sbig_pulse(&pulse);
	    printf("sbig_pulse returned %d\n", ret);
	    if (ret < 0)
		printf("sbig error: %s\n", sbig_show_error(ret));
	    continue;
	}
	if (strcmp(buffer, "pulse\n") == 0) {
	    struct sbig_pulse  pulse;
	    pulse.nmbr_pulses = geti("nmbr of pulses: ", 0);
	    pulse.pulse_width = geti("pulse width in us [>= 9]: ", 0);
	    pulse.pulse_interval = geti("pulse interval [>= 29+width]: ", 0);
	    sync();
	    ret = sbig_pulse(&pulse);
	    printf("sbig_pulse returned %d\n", ret);
	    if (ret < 0)
		printf("sbig error: %s\n", sbig_show_error(ret));
	    continue;
	}
	if (strcmp(buffer, "cool\n") == 0) {
	    double  f;
	    struct sbig_cool  cool;
	    cool.regulation = geti("Cooling [0 Off, 1 *On, 2 DirectDrive]: ", 1);
	    cool.direct_drive = 0;
	    if (cool.regulation == 2)
		cool.direct_drive = geti("Power [0..255, default 200]: ", 200);
	    else if (cool.regulation == 1) {
		f = getf("Enter temperature in deg C [default 0.0]: ", 0.0);
		i = 10.0*f + 0.5;
		cool.temperature = i;
	    }
	    sync();
	    ret = sbig_set_cooling(&cool);
	    printf("sbig_set_cooling() returned %d\n", ret);
	    if (ret < 0)
		printf("sbig error: %s\n", sbig_show_error(ret));
	}
    }
    exit(0);
}
